#version 430
#extension GL_ARB_separate_shader_objects : enable










struct SComputeInput
{
	uvec3 GlobalThreadID; uvec3 LocalThreadID; uvec3 GroupID;};

layout( std140) uniform USelectionInfo
{
	uvec4 PositionsOffset;
	uvec4 RadiusOffset;
	uvec4 EnabledOffset;
	vec4 RayOrigin;
	vec4 RayDirection;
	mat4 Planes;
	uint Mode;
} SelectionInfo;
layout(std430 ) readonly buffer BStream
{
	uint Stream[];
} Stream;
layout(std430 ) readonly buffer BStreamInfo
{
	uint StreamInfo[];
} StreamInfo;
layout(std430 ) buffer BSelection
{
	uint Selection[];
} Selection;



void ComputeMain( in SComputeInput  cInput  )
{



	const uint threadId = cInput.GlobalThreadID.x;
	if (threadId != 0)
		return;

	const uint count =  StreamInfo.StreamInfo [0] ;
	if (count == 0)
		return;

	const float DebugColorNo = 0.f;
	const float DebugColorWhite = 2.0f;
	const float DebugColorRed = 3.0f;

	const bool advanceNext =  SelectionInfo.Mode  == 8;
	const bool modeReset =  SelectionInfo.Mode  == 10;

	uint oldParticleId = uint(-1);

	for (uint i = 0; i < count; ++i)
	{
		const float sel =  uintBitsToFloat(Selection.Selection [ i ]) ;
		if (sel == DebugColorRed)
		{
			oldParticleId = i;
			break;
		}
	}

	uint newParticleId = oldParticleId;

	if (modeReset)
	{
		newParticleId = uint(-1);
	}
	else if (oldParticleId != uint(-1))
	{
		while (true)
		{
			const float sel =  uintBitsToFloat(Selection.Selection [ newParticleId ]) ;
			if (sel == DebugColorWhite)
				break;

			if (advanceNext)
			{
				if (++newParticleId == count)
					newParticleId = 0;
			}
			else
			{
				if (newParticleId-- == 0)
					newParticleId = count - 1;
			}
			if (newParticleId == oldParticleId)
				break;
		}
	}
	else
	{
		for (uint i = 0; i < count; ++i)
		{
			const float sel =  uintBitsToFloat(Selection.Selection [ i ]) ;
			if (sel != DebugColorNo)
			{
				newParticleId = i;
				break;
			}
		}
	}

	if (oldParticleId != uint(-1) && newParticleId != oldParticleId)
	{
		Selection.Selection [ oldParticleId ] = floatBitsToUint(DebugColorWhite) ; ;
	}

	if (newParticleId != uint(-1))
	{
		Selection.Selection [ newParticleId ] = floatBitsToUint(DebugColorRed) ; ;
	}
}


layout(local_size_x = 128, local_size_y = 1, local_size_z = 1) in;
void main()
{
	SComputeInput cInput;
	cInput.GlobalThreadID = gl_GlobalInvocationID;
	cInput.LocalThreadID = gl_LocalInvocationID;
	cInput.GroupID = gl_WorkGroupID;
	ComputeMain(cInput);
}
