#version 430
#extension GL_ARB_separate_shader_objects : enable










struct SComputeInput
{
	uvec3 GlobalThreadID; uvec3 LocalThreadID; uvec3 GroupID;};

layout(std430 ) readonly buffer BInCounts
{
	uint InCounts[];
} InCounts;
layout(std430 ) readonly buffer BInKeys
{
	uint InKeys[];
} InKeys;
layout(std430 ) readonly buffer BInIndirection
{
	uint InIndirection[];
} InIndirection;
layout(std430 ) buffer BOutKeys
{
	uint OutKeys[];
} OutKeys;
layout(std430 ) buffer BOutIndirection
{
	uint OutIndirection[];
} OutIndirection;

layout( std140) uniform UInfo
{
	uint CurrentBit;
	uint GroupCount;
} Info;

shared uint localCounts[256];

void ComputeMain( in SComputeInput  cInput  )
{
	const uint globalThreadID = cInput.GlobalThreadID.x;
	const uint localThreadID = cInput.LocalThreadID.x;
	const uint groupID = cInput.GroupID.x;

	const uint currentBit =  Info.CurrentBit ;
	const uint groupCount =  Info.GroupCount ;


	const uvec4 fullKeys =  uvec4( InKeys.InKeys [ globalThreadID * 4 ], InKeys.InKeys [( globalThreadID * 4 ) + 1], InKeys.InKeys [( globalThreadID * 4 ) + 2], InKeys.InKeys ( globalThreadID * 4 ) + 3]) ;
	uvec2 keyBits;
	keyBits.x = ((currentBit < 32 ? fullKeys.y : fullKeys.x) >> (currentBit % 32)) & 1;
	keyBits.y = ((currentBit < 32 ? fullKeys.w : fullKeys.z) >> (currentBit % 32)) & 1;
	const uvec2 startValues = 1 - keyBits;
	localCounts [localThreadID * 2 + 0] = startValues.x;
	localCounts [localThreadID * 2 + 1] = startValues.y;

	uint stride = 1;
	for (uint d =  128 ; d > 0; d /= 2)
	{
		groupMemoryBarrier(); barrier(); ;
		if (localThreadID < d)
		{
			const uint ai = (localThreadID * 2 + 1) * stride - 1;
			const uint bi = (localThreadID * 2 + 2) * stride - 1;
			localCounts [bi] +=  localCounts [ai];
		}
		stride *= 2;
	}

	groupMemoryBarrier(); barrier(); ;

	if (localThreadID == 0)
	{
		localCounts [ 128  * 2- 1] = 0;
	}

	for (uint d = 1; d <  128  * 2; d *= 2)
	{
		stride /= 2;
		groupMemoryBarrier(); barrier(); ;
		if (localThreadID < d)
		{
			const uint ai = (localThreadID * 2 + 1) * stride - 1;
			const uint bi = (localThreadID * 2 + 2) * stride - 1;
			const uint t =  localCounts [ai];
			localCounts [ai] =  localCounts [bi];
			localCounts [bi] += t;
		}
	}

	groupMemoryBarrier(); barrier(); ;

	const uint globalTotalSum =  InCounts.InCounts [ groupCount - 1 ] ;
	uint lastTotal;

	if (groupID == 0)
		lastTotal = 0;
	else
		lastTotal =  InCounts.InCounts [ groupID - 1 ] ;

	const uvec2 sums = uvec2( localCounts [localThreadID * 2 + 0],  localCounts [localThreadID * 2 + 1]) + lastTotal;
	const uvec2 k01 = uvec2(0, 1);
	const uvec2 dstIf1 = (globalThreadID * 2 + k01) - sums + globalTotalSum;
	const uvec2 mask = uvec2(keyBits.x != 0 ? ~0 : 0, keyBits.y != 0 ? ~0 : 0);
	const uvec2 dsts = (mask & dstIf1) | (~mask & sums);

	const uvec2 dstsAddr =  dsts ;


	OutKeys.OutKeys [dstsAddr.x * 2] = fullKeys.xy.x; OutKeys.OutKeys [(dstsAddr.x * 2) + 1] = fullKeys.xy.y ;
	OutKeys.OutKeys [dstsAddr.y * 2] = fullKeys.zw.x; OutKeys.OutKeys [(dstsAddr.y * 2) + 1] = fullKeys.zw.y ;
	const uvec2 inValues =  uvec2( InIndirection.InIndirection [ globalThreadID * 2 ], InIndirection.InIndirection [( globalThreadID * 2 ) + 1]) ;

	OutIndirection.OutIndirection [dstsAddr.x] = inValues.x; ;
	OutIndirection.OutIndirection [dstsAddr.y] = inValues.y; ;
}

layout(local_size_x = 128, local_size_y = 1, local_size_z = 1) in;
void main()
{
	SComputeInput cInput;
	cInput.GlobalThreadID = gl_GlobalInvocationID;
	cInput.LocalThreadID = gl_LocalInvocationID;
	cInput.GroupID = gl_WorkGroupID;
	ComputeMain(cInput);
}
